#= Metasequoiaf[^SAXp[T
# 'mqo/parser.rb'

require 'strscan.so'

module Metasequoia
   class FormatError < StandardError
   end

   class Parser
      MQO_Float = /-?(?:\d+\.)?\d+\s*/
      MQO_Int = /-?\d+\s*/
      MQO_Name = /\"([^\"]*)\"\s*/
      MQO_Attr = /(\w+)\s+/
      MQO_Param = /(\w+)\(([^\)]+)\)\s*/
      MQO_ParamStart = /(\w+)\(/
      MQO_ParamEnd = /\)\s*/
      MQO_Vector = /(#{MQO_Float})\s+(#{MQO_Float})\s+(#{MQO_Float})/
      MQO_RGBA = /(#{MQO_Float})\s+(#{MQO_Float})\s+(#{MQO_Float})\s+(#{MQO_Float})/
      MQO_ChunkStart = /(\w+)\s+(?:(.+)\s*)?\{/
      MQO_ChunkEnd = /\}/
      MQO_Eof = /Eof/i
      MQO_Spaces = /\s+/
      MQO_Face = /(\d)\s+V\(/
      MQO_Vertex = MQO_Vector
      MQO_Material = /#{MQO_Name}\s+/
      MQO_Weit = /(\d+)\s+(#{MQO_Float})/
      MQO_BVector = /Vector\s+(\d+)\s+\[(\d+)\]/i
      MQO_BackImage = /#{MQO_Name}#{MQO_RGBA}/

      def initialize(filename)
         @file = filename
      end

      # p[X̃C[`
      def parse
         open(@file, 'rb'){|io|
            raise FormatError unless io.gets =~ /^Metasequoia Document/
            raise FormatError unless io.gets =~ /^Format Text Ver (#{MQO_Float})/
            @level = 0
            @root = @current = document_start($1.to_f)
            @parent = [@root]
            io.each{|line|
               @scanner = StringScanner.new(line)
               @lineno = io.lineno
               @scanner.skip MQO_Spaces
               next if @scanner.eos?

               case @level
               when 0 #----------------------------------------------level 0
                  if @scanner.scan MQO_ChunkStart
                     @parent.push @current
                     case @scanner[1].downcase.intern
                     when :scene
                        @current = scene_chunk_start
                     when :backimage
                        @current = backimage_chunk_start
                     when :material
                        @current = material_chunk_start(@scanner[2].to_i)
                     when :object
                        MQO_Name =~ @scanner[2]
                        @current = object_chunk_start($1)
                     when :blob
                        @current = blob_chunk_start(@scanner[2])
                     else
                        @current = chunk_start(@scanner[1], @scanner[2])
                     end
                     @level = 1
                     next

                  elsif @scanner.scan MQO_Eof	# MQO_AttrɂƂ܂Ȃ
                     print "Eof\n" if $Debug
                     break
                  end


               when 1 #----------------------------------------------level 1
                  if @scanner.scan MQO_ChunkStart
                     @parent.push @current
                     case @scanner[1].downcase.intern
                     when :vertex, :bvertex
                        @current = vertex_chunk_start(@scanner[2].to_i)
                     when :face
                        @current = face_chunk_start(@scanner[2].to_i)
                     when :vertexattr
                        @current = vertexattr_chunk_start
                     else
                        @current = chunk_start(@scanner[1], @scanner[2])
                     end
                     @level = 2
                     next

                  elsif @scanner.scan MQO_Material
                     parse_material(@scanner[1])
                     next
                  end


               when 2 #----------------------------------------------level 2
                  if @scanner.scan MQO_Face
                     parse_face(@scanner[1].to_i)
                     next

                  elsif @scanner.scan MQO_Vertex
                     vertex_value(@scanner[1].to_f, @scanner[2].to_f, @scanner[3].to_f)
                     next

                  elsif @scanner.scan MQO_ChunkStart
                     @parent.push @current
                     case @scanner[1].downcase.intern
                     when :weit
                        @current = weit_chunk_start
                     else
                        @current = chunk_start(@scanner[1], @scanner[2])
                     end
                     @level = 3
                     next

                  elsif @scanner.scan MQO_BVector
                     vnum = @scanner[1].to_i
                     byte = @scanner[2].to_i
                     raise FormatError if byte != vnum * 12
                     tmp = io.read(byte)
                     vnum.times{|i|
                        v = tmp[12*i, 12].unpack('e3')
                        vertex_value(v[0], v[1], v[2])
                     }
                     next
                  end
               end

               #----------------------------------------------level any
               if @scanner.scan MQO_ChunkStart
                  @parent.push @current
                  @current = chunk_start(@scanner[1], @scanner[2])
                  @level += 1

               elsif @scanner.scan MQO_ChunkEnd
                  @level -= 1
                  chunk_end
                  @current = @parent.pop

               elsif @scanner.scan MQO_Weit
                  weit_value(@scanner[1].to_i, @scanner[2].to_f)

               elsif @scanner.scan MQO_Attr
                  parse_attr(@scanner[1])

               else
                  if $Debug
                     print @scanner.string,"\n"
                  end
               end
            } rescue begin STDERR.puts "\"#@file\":#{@lineno}\nError!\n#{@scanner.string}\n#$!\n#{$@.join("\n")}"; return nil end
            document_end
         }
      end

=begin
ʂȃGg̃p[X
{I parse_* \bh͍Ē`̕KvȂB
=end
      def parse_attr(atr)
         if @scanner.scan MQO_Vector
            attr_vector_value(atr, @scanner[1].to_f, @scanner[2].to_f, @scanner[3].to_f)
         elsif @scanner.scan MQO_BackImage
            backimage_value(atr, @scanner[1], [@scanner[2].to_f, @scanner[3].to_f, @scanner[4].to_f, @scanner[5].to_f])
         elsif @scanner.scan MQO_Float
            attr_float_value(atr, @scanner[0].to_f)
         elsif @scanner.scan MQO_Int
            attr_int_value(atr, @scanner[0].to_i)
         else
            STDERR.puts "\"#@file\":#{@lineno}\nWarning! Unknown attribute '#{atr}' of #{@current.class}\n#{@scanner.string}"
         end
      end
      def parse_face(vnum)
         case vnum
         when 2
            vtx = [
            @scanner.scan(MQO_Int).to_i,
            @scanner.scan(MQO_Int).to_i
            ]
         when 3
            vtx = [
            @scanner.scan(MQO_Int).to_i,
            @scanner.scan(MQO_Int).to_i,
            @scanner.scan(MQO_Int).to_i
            ]
         when 4
            vtx = [
            @scanner.scan(MQO_Int).to_i,
            @scanner.scan(MQO_Int).to_i,
            @scanner.scan(MQO_Int).to_i,
            @scanner.scan(MQO_Int).to_i
            ]
         else
            raise FormatError
         end
         raise FormatError if @scanner.skip(MQO_ParamEnd).nil?
         mat_id = nil
         uv_ary = nil
         vcol = nil
         while !@scanner.eos?
            break if @scanner.scan(MQO_ParamStart).nil?
            case @scanner[1].intern
            when :M
               raise FormatError if @scanner.scan(MQO_Int).nil?
               mat_id = @scanner[0].to_i
            when :UV
               case vnum
               when 2
                  uv_ary = [
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f],
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f]
                  ]
               when 3
                  uv_ary = [
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f],
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f],
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f]
                  ]
               when 4
                  uv_ary = [
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f],
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f],
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f],
                  [@scanner.scan(MQO_Float).to_f, @scanner.scan(MQO_Float).to_f]
                  ]
               end
            when :COL
               case vnum
               when 2
                  vcol = [
                  @scanner.scan(MQO_Int).to_i,
                  @scanner.scan(MQO_Int).to_i
                  ]
               when 3
                  vcol = [
                  @scanner.scan(MQO_Int).to_i,
                  @scanner.scan(MQO_Int).to_i,
                  @scanner.scan(MQO_Int).to_i
                  ]
               when 4
                  vcol = [
                  @scanner.scan(MQO_Int).to_i,
                  @scanner.scan(MQO_Int).to_i,
                  @scanner.scan(MQO_Int).to_i,
                  @scanner.scan(MQO_Int).to_i
                  ]
               end
            else
               break
            end
            raise FormatError if @scanner.skip(MQO_ParamEnd).nil?
         end
         face_value(vnum, vtx, mat_id, uv_ary, vcol)
      end
      def parse_material(mat_name)
         mat_hash = Hash.new
         rgba = nil
         while !@scanner.eos?
            break if @scanner.scan(MQO_Param).nil?
            case @scanner[1]
            when 'col'
               MQO_RGBA =~ @scanner[2]
               rgba = [$1.to_f, $2.to_f, $3.to_f, $4.to_f]
            else
               mat_hash[@scanner[1]] = @scanner[2]
            end
         end
         material_value(mat_name, rgba, mat_hash)
      end

=begin
pNXł͈ȉ̃\bhĒ`ĎgB
=end
      # DOMo͂ꍇ̓[gIuWFNgo͂
      def document_start(ver)
         if $Debug
            print "Metasequoia Document\nFormat Text Ver #{ver}\n\n"
         end
         return nil
      end
      def document_end
      end
      # m̃`N̏
      def chunk_start(chunk, opt)
         if $Debug
            print "#{' '*@level}#{chunk} #{opt} {\n"
         end
         return nil
      end
      # DOMo͂ꍇ̓V[IuWFNgo͂
      def scene_chunk_start
         if $Debug
            print "#{' '*@level}Scene \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͉GIuWFNgo͂
      def backimage_chunk_start
         if $Debug
            print "#{' '*@level}BackImage \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͍ގXgIuWFNgo͂
      def material_chunk_start(mat_num)
         if $Debug
            print "#{' '*@level}Material #{mat_num} \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ̓fIuWFNgo͂
      def object_chunk_start(obj_name)
         if $Debug
            print "#{' '*@level}Object \"#{obj_name}\" \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͔SyIuWFNgo͂
      def blob_chunk_start(opt)
         if $Debug
            print "#{' '*@level}Blob #{opt} \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͒_XgIuWFNgo͂
      def vertex_chunk_start(vnum)
         if $Debug
            print "#{' '*@level}vertex #{vnum} \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͒_XgIuWFNgo͂
      def vertexattr_chunk_start()
         if $Debug
            print "#{' '*@level}vertexattr \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͒_d݃XgIuWFNgo͂
      def weit_chunk_start()
         if $Debug
            print "#{' '*@level}weit \{\n"
         end
         return nil
      end
      # DOMo͂ꍇ͖ʃXgIuWFNgo͂
      def face_chunk_start(fnum)
         if $Debug
            print "#{' '*@level}face #{fnum} \{\n"
         end
         return nil
      end
      # `N̏ICxg
      def chunk_end
         if $Debug
            print "#{' '*@level}}\n"
         end
      end
      # _f[^
      def vertex_value(x, y, z)
         if $Debug
            print "#{' '*@level}(#{x}, #{y}, #{z})\n"
         end
      end
      # _̏d
      def weit_value(vid, weit)
         if $Debug
            print "#{' '*@level}#{vid} #{weit}\n"
         end
      end
      # ʃf[^
      def face_value(vnum, vtx, mat_id, uv_ary, vcol)
         if $Debug
            print "#{' '*@level}#{vnum} V(#{vtx.join(' ')})"
            print " M(#{mat_id})" unless mat_id.nil?
            print " UV(#{uv_ary.join(' ')})" unless uv_ary.nil?
            print " COL(#{vcol.join(' ')})" unless vcol.nil?
            print "\n"
         end
      end
      # ގf[^
      def material_value(mat_name, rgba, mat_hash)
         if $Debug
            print "#{' '*@level}\"#{mat_name}\" col(#{rgba.join(' ')})"
            mat_hash.each{|key,val|
               print " #{key}(#{val})"
            }
            print "\n"
         end
      end
      # Gf[^
      def backimage_value(view, file, opt)
         if $Debug
            print "#{' '*@level}#{view} \"#{file}\" #{opt.join(' ')}}\n"
         end
      end
      # xNg^̑f[^
      def attr_vector_value(atr, x, y, z)
         if $Debug
            print "#{' '*@level}#{atr} #{x} #{y} #{z}\n"
         end
      end
      # Float^̑f[^
      def attr_float_value(atr, val)
         if $Debug
            print "#{' '*@level}#{atr} #{val}\n"
         end
      end
      # Int^̑f[^
      def attr_int_value(atr, val)
         if $Debug
            print "#{' '*@level}#{atr} #{val}\n"
         end
      end
   end
end

#= PƂŌĂяoꂽꍇ̂ݗLȏB
# CuƂČĂяoꂽꍇɂ͖B
if $0 == __FILE__
   #$Debug = true
   module Metasequoia
      class Loader < Parser
         def load
            parse
            return @root
         end
      end

      def self.load(filename)
         raise FormatError unless filename =~ /\.mqo$/i
         mqoLD = Loader.new(filename)
         return mqoLD.load
      end
   end

   print "MQO file loading ..."
   ptime = Process.times.utime
   Metasequoia::load('test.mqo')
   print "complete! (Elapsed time: #{Process.times.utime - ptime}s)\n"
end
