#! /usr/bin/ruby
#= MetasequoiaMarbleCLAYϊXNvg

$Scaling = 0.1
module MetaCLAY
   class Vector
      def initialize(x,y,z)
         @pt = [x*$Scaling, y*$Scaling, z*$Scaling]
      end
      def to_s
         @pt.join(',')
      end
   end
   class UVs
      attr_reader :pt
      def initialize(d,u,v)
         @pt = [d, u, v]
      end
      def to_s
         @pt.join(',')
      end
   end

   def self.mqo2mcl(mqo_file)
      mcl_file = mqo_file.sub(/\.mqo$/i,'.mcl')
      open(mqo_file, 'rb'){|mqo|
         open(mcl_file, 'w'){|mcl|
            return nil unless mqo.gets =~ /^Metasequoia Document/
            return nil unless mqo.gets =~ /^Format Text Ver 1.0/

            mcl.print <<-EOH
Header {
  CreatedBy mqo2mcl alpha-3
}
            EOH
            loop{
               ln = mqo.gets
               case ln
               when /^Scene \{/
                  self.m2c_scene(mqo, mcl)
               when /^Material (\d+) \{/
                  self.m2c_mat(mqo, mcl)
               when /^Object \"(.*)\" \{/
                  self.m2c_obj(mqo, mcl, $1)
               when /^Eof/
                  return true
               end
            }
         }
      }
   end

   def self.m2c_scene(mqo, mcl)
      pos = nil
      lookat = nil
      amb = nil
      head = nil
      pich = nil
      zoom2 = nil
      loop{
         case mqo.gets
         when /(pos|lookat) (-?\d+\.\d+) (-?\d+\.\d+) (-?\d+\.\d+)/
            eval("#{$1} = Vector.new(#{$2}, #{$3}, #{$4})")
         when /amb (-?\d+\.\d+) (-?\d+\.\d+) (-?\d+\.\d+)/
            amb = "#{$1} #{$2} #{$3}"
         when /(head|pich|zoom2) (-?\d+\.\d+)/
            eval("#{$1} = #{$2}")
         when /ortho ([01])/
            ortho = $1.to_i
         when /^\}/
            mcl.print <<-EOD
 Document {
   GridWidth 1
   LargeGrid 5
   GridSnap 0
   Rev 2
   DrawLockedVertex 0
   Vx.0 50
   Hy.0 50
   Rz.0 100
   ViewCenter.x.0 0
   ViewCenter.y.0 0
   ViewCenter.z.0 0
   Vx.1 0
   Hy.1 0
   Rz.1 0
   ViewCenter.x.1 0
   ViewCenter.y.1 0
   ViewCenter.z.1 0
   Vx.2 0
   Hy.2 0
   Rz.2 0
   ViewCenter.x.2 0
   ViewCenter.y.2 0
   ViewCenter.z.2 0
   Vx.3 0
   Hy.3 0
   Rz.3 0
   ViewCenter.x.3 0
   ViewCenter.y.3 0
   ViewCenter.z.3 0
 }
            EOD
            return true
         end
      }
   end

   def self.m2c_mat(mqo, mcl)
      tex = nil
      alpha = nil
      col = nil
      mcl.print "Materials {\n"
      loop{
         case mqo.gets
         when /"([^"]+)" (.+)/
            mcl.print "  Material {\n    name \"#{$1}\"\n"
            $2.scan(/(\w+)\(([^)]+)\)/){|matched|
               case matched[0]
               when 'col'
                  matched[1] =~ /(-?\d+\.\d+) (-?\d+\.\d+) (-?\d+\.\d+) (-?\d+\.\d+)/
                  r, g, b = ($1.to_f*255).to_i, ($2.to_f*255).to_i, ($3.to_f*255).to_i
                  mcl.print "    Material {\n      name Diffuse\n      col3 #{r} #{g} #{b}\n    }\n"
               end
            }
            mcl.print "  }\n"
         when /\}/
            mcl.print "}\n"
            return true
         end
      }
   end

   def self.m2c_obj(mqo, mcl, objname)
      visible = 1
      locking = 0
      facet = nil
      mcl.print "Object {\n   Structure {\n      Name \"#{objname}\"\n"
      loop{
         case mqo.gets
         when /visible (\d+)/
            visible = ($1.to_i > 0) ? 1 : 0
         when /locking (\d+)/
            locking = ($1.to_i > 0) ? 1 : 0
         when /facet (\d+\.\d+)/
            facet = $1.to_f
         when /vertex (\d+) \{/
            mcl.print <<-EOM
      Stat #{visible} #{locking}
      Mesh "Mesh-0" 0 #{visible} #{locking}
   }
  Vertexs {
            EOM
            vnum = $1.to_i
            vnum.times{|i|
               mqo.gets =~ /(-?\d+\.\d+) (-?\d+\.\d+) (-?\d+\.\d+)/
               pt = Vector.new($3.to_f,$1.to_f,$2.to_f)
               mcl.print "    0 #{pt} 0 -1 0\n"
            }
            while !(mqo.gets =~ /\}/)	# of vertex
            end
            mcl.print "  }\n"
         when /BVertex (\d+) \{/
            mcl.print <<-EOM
      Stat #{visible} #{locking}
      Mesh "Mesh-0" 0 1 0
   }
  Vertexs {
            EOM
            vnum = $1.to_i
            mqo.gets =~ /Vector (\d+) \[(\d+)\]/
            vdat = mqo.read($2.to_i)
            vnum.times{|i|
               v = vdat[12*i, 12].unpack('e3')
               mcl.print "    0 #{v[1]},#{v[2]},#{v[0]} 0 -1 0\n"
            }
            while !(mqo.gets =~ /\}/)	# of Vector
            end
            while !(mqo.gets =~ /\}/)	# of BVertex
            end
            mcl.print "  }\n"
         when /face (\d+) \{/
            poly_block = []
            uv_block = []
            uv_no = 0
            fnum = $1.to_i
            fnum.times{|i|
               mqo.gets =~ /(\d) V\(([^)]+)\)( M\((\d+)\))?( UV\(([^)]+)\))?/
               fv = $1.to_i
               if fv == 3 || fv == 4
                  pt = $2
                  mat = $4 || '-1'
                  uv = $6
                  unless uv.nil?
                     tmp = uv.scan(/(-?\d\.\d+) (-?\d\.\d+)/)
                     tmp.map!{|x| UVs.new(0, x[0].to_f, 1.0-x[1].to_f)}
                     uv_block += tmp
                  end
                  tmp = '    0 '
                  tmp += pt.scan(/\d+/).join(',')
                  tmp += ' 0 0 -1 '
                  if uv.nil?
                     tmp += Array.new(fv,-1).join(',')
                  else
                     tmp += (uv_no..(uv_no+fv-1)).to_a.join(',')
                  end
                  tmp += " #{mat}\n"
                  poly_block << tmp
                  uv_no += fv unless uv.nil?
               end
            }
            while !(mqo.gets =~ /\}/)	# of face
            end
            if uv_no > 0
               mcl.print "  UVs {\n"
               uv_block.each{|x|
                  mcl.print "    0 #{x} 0 -1\n"
               }
               mcl.print "  }\n"
            end
            mcl.print "  Polygons {\n"
            mcl.print poly_block
            mcl.print "  }\n"
         when /\}/
            mcl.print "}\n"
            return true
         end
      }
   end

end

#= PƂŌĂяoꂽꍇ̂ݗLȏB
# CuƂČĂяoꂽꍇɂ͖B
if $0 == __FILE__
   require 'optparse'
   ARGV.options{|opts|
      opts.banner = "Usage: #{$0} [options] mqo_file ...\n"

      opts.on_head("normal options:")

      opts.on("--scaling=SCALING", Float, ":float [0.1]") {|$Scaling|}

      opts.on_tail # separater
      opts.on_tail("common options:")
      opts.on_tail("-?", "--help", "show this message") {puts opts; exit}
      opts.parse!
   } or (puts "\n", ARGV.options; exit 1)

   ARGV.reject!{|x|
      !(/\.mqo$/i =~ x)
   }
   (puts ARGV.options; exit) if ARGV.empty?
   ARGV.options = nil	# no more parse

   ARGV.each{|fn|
      MetaCLAY::mqo2mcl(fn)
   }
end
