#! /usr/bin/ruby
#= Ruby/MetaRed: MetasequoiaRedqueen2ϊXNvg(mqo2rq2.rb)

$KCODE = 'Shift_JIS'
#$Debug = true

require 'mqo/loader'
require 'rq2/dom'

module Metasequoia
   FMT_Vertex   = '    <vertex x="%f" y="%f" z="%f" />'
   FMT_VertexUV = '    <vertex x="%f" y="%f" z="%f" u="%f" v="%f" />'
   FMT_Triangle = '    <triangle id0="%d" id1="%d" id2="%d" />'
   class FaceChunk
      attr_accessor :vertices
      attr_reader :part
      def initialize
         @part = Hash.new
      end
   end
   class VertexChunk
      def <<(data)
         @chunk << data
      end
   end
   class Vector
      @@FMT_Vector = '<%s x="%f" y="%f" z="%f" />'
      def rq2_tag(tag, scale = $scale)
         sprintf(@@FMT_Vector, tag, @x*scale, -@z*scale, @y*scale)
      end
      def to_rq2(name = 'vector', scale = $scale)
         v = Redqueen2::Vector.new(@x*scale, -@z*scale, @y*scale)
         v.name = name
         return v
      end
   end
   class Scene
      attr_accessor :backimages
      def fputs(opts, out = STDOUT)
         self.fputs_camera(opts, out)
         self.fputs_renderer(opts, out)
         self.fputs_sky(opts, out)
      end

      def fputs_camera(opts, out = STDOUT)
         # Camera
         location, target, eye, upvector, fov = camera_info
         eye = eye.to_rq2('eye',1)
         upvector = upvector.to_rq2('upvector', 1)

         cam = Redqueen2::Camera.new('pinhole', opts[:width], opts[:height], opts[:oversampling], fov, opts[:blur])
         cam << location.to_rq2('location', opts[:camera_scaling])
         cam << target.to_rq2('target', opts[:camera_scaling])
         cam << upvector
         cam.fputs(out, 1)

         # Cgo
         d = upvector - eye
         # 牺45xƂ炷Cg
         case opts[:default_lights]
         when true
            Redqueen2::ParallelLight.new(Redqueen2::Color.new(0.6, 0.6, 0.6), d, 50000, 1).fputs(out, 1)
         when Proc
            tmp = opts[:default_lights].call(eye, upvector)
            case tmp
            when Array
               tmp.each{|x|
                  x.fputs(out, 1) if x.is_a? Redqueen2::Light
               }
            when Redqueen2::Light
               tmp.fputs(out, 1)
            end
         end
      end

      def fputs_renderer(opts, out = STDOUT)
         # Renderer
         rend = Redqueen2::Renderer.new(opts[:tracedepth], opts[:gisample], opts[:photonsearchradius], opts[:diffusealbedo], opts[:diffusedepth])
         coff = opts[:cutoff] || Redqueen2::Color.new(0.3, 0.3, 0.3)
         coff.name = 'cutoff'
         rend << coff
         case opts[:ambient]
         when :use_mqo
            amb = Redqueen2::Color.new(@amb.x, @amb.y, @amb.z, 'ambient')
         when Redqueen2::Color
            amb = opts[:ambient]
            amb.name = 'ambient'
         else
            amb = Redqueen2::Color.new(0, 0, 0, 'ambient')
         end
         rend << amb
         rend.fputs(out, 1)
      end

      def fputs_sky(opts, out = STDOUT)
         # Sky
         sky = Redqueen2::Sky.new(opts[:bgtype])
         sky['date'] = opts[:date]
         sky['time'] = opts[:time]
         sky['index'] = opts[:bgindex]
         sky['meridian'] = opts[:meridian]
         sky['turbidity'] = opts[:turbidity]
         sky['latitude'] = opts[:latitude]
         sky['longitude'] = opts[:longitude]
         sky['photon'] = opts[:bgphoton]
         unless opts[:zenith].nil?
            opts[:zenith].name = 'zenith'
            sky << opts[:zenith]
         end
         unless opts[:north].nil?
            opts[:north].name = 'north'
            sky << opts[:north]
         end
         unless opts[:bgtiling].nil?
            opts[:bgtiling].name = 'tiling'
            sky << opts[:bgtiling]
         end
         sky << (opts[:bgcolor] || Redqueen2::Color.new(@amb.x, @amb.y, @amb.z))
         hdri = @backimages['pers'].file unless @backimages.nil? || @backimages['pers'].nil?
         hdri = opts[:hdri]
         sky << Redqueen2::Image.new(hdri.gsub(/\\/,'/')) unless hdri.nil?
         sky.fputs(out, 1)
      end
   end

   class Object
      attr_accessor :root
      def fputs(opts, out = STDOUT)
         return if @visible == 0

         puts " obj: #{@name}" if $Debug
         out.puts " <object name=\"#{@name}\" smoothing=\"#{@facet}\" >"
         part = @faces.part
         vtx = Hash.new
         part_opt = ''
         part_opt += 'castshadow="false" ' if @locking == 1
         part_opt += 'primaryunvisible="true" ' if @name =~ /^@/
         part_opt += 'throughphoton="true" ' if @name =~ /!$/
         part.each{|key, f_ary|
            mat_name = key == :default ? 'default_material' : @root.materials[key].name
            puts "  part: #{mat_name}" if $Debug
            tmp = f_ary.flatten
            tmp.uniq!
            out.print "  <part material=\"#{mat_name}\" #{part_opt}>\n"
            out.print "   <vertices>\n"
            tmp.each_with_index{|x,i|
               out.puts x
               vtx[x] = i
            }
            out.print "   </vertices>\n"
            out.print "   <triangles>\n"
            f_ary.each{|x|
               out.printf(FMT_Triangle, vtx[x[0]], vtx[x[1]], vtx[x[2]])
               out.puts
            }
            out.print "   </triangles>\n"
            out.print "  </part>\n"
            tmp.clear
            vtx.clear
         }
         out.puts " </object>"
      end
   end

   class MaterialChunk
      def fputs(opts, out = STDOUT)
         self.each{|m|
            m.fputs(opts, out)
         }
      end
   end

   class Material
      def fputs(opts, out = STDOUT)
         out.print " <material name=\"#{@name}\" "
         ref_idx = nil
         ref_idx = 1 if @color.alpha < 1
         ref_idx = $1 if @name =~ /\(@((?:\d+\.)?\d+)\)/
         out.print "refractionindex=\"#{ref_idx}\" " unless ref_idx.nil?
         out.print ">\n"
         fputs_mat_diffusion(opts, out)
         fputs_mat_transparency(opts, out)
         fputs_mat_trimming(opts, out)
         fputs_mat_bumpiness(opts, out)

         fputs_mat_sss(opts, out)

         fputs_mat_emission(opts, out) unless opts[:no_emi_to_emission]
         fputs_mat_glossiness(opts, out) unless opts[:no_amb_to_glossiness]
         fputs_mat_reflection(opts, out) unless opts[:no_dif_to_reflection]
         fputs_mat_specularity(opts, out) unless opts[:no_spc_to_specularity]
         out.print " </material>\n"
      end

      def fputs_mat_trimming(opts, out)
         out.print "  <trimming>\n   <image name=\"#{@aplane.downcase.gsub(/\\/,'/')}\" />\n  </trimming>\n" unless @aplane.nil?
      end

      def fputs_mat_bumpiness(opts, out)
         out.print "  <bumpiness>\n   <image name=\"#{@bump.downcase.gsub(/\\/,'/')}\" />\n  </bumpiness>\n" unless @bump.nil?
      end

      def fputs_mat_sss(opts, out)
         # <absorb r="0" g="0" b="0" />
         # <scatter r="0" g="0" b="0" />
      end

      def fputs_mat_diffusion(opt, out)
         return if @name =~ /\[\w*_\w*\]/	# mat01[_???]Ƃdiffusion͏o͂Ȃ

         out.printf("  <diffusion>\n   <color r=\"%f\" g=\"%f\" b=\"%f\" />\n", @color.r, @color.g, @color.b)
         out.print "   <image name=\"#{@tex.downcase.gsub(/\\/,'/')}\" />\n" unless @tex.nil?
         out.print "  </diffusion>\n"
      end

      def fputs_mat_transparency(opts, out)
         a = 1.0 - @color.alpha
         @name =~ /\[\w*(t|T)\w*\]/
         fmt = "  <transparency>\n   <color r=\"%f\" g=\"%f\" b=\"%f\" />\n  </transparency>\n"
         case $1
         when 't'
            out.printf(fmt, @color.r, @color.g, @color.b)
         when 'T'
            return if a == 0
            out.printf(fmt, @color.r*a, @color.g*a, @color.b*a)
         else
            return if a == 0
            out.printf(fmt, a, a, a)
         end
      end

      def fputs_mat_reflection(opts, out)
         a = 1.0 - @dif
         @name =~ /\[\w*(r|R)\w*\]/
         fmt = "  <reflection>\n   <color r=\"%f\" g=\"%f\" b=\"%f\" />\n  </reflection>\n"
         case $1
         when 'r'
            out.printf(fmt, @color.r, @color.g, @color.b)
         when 'R'
            return if a == 0
            out.printf(fmt, @color.r*a, @color.g*a, @color.b*a)
         else
            return if a == 0
            out.printf(fmt, a, a, a)
         end
      end

      def fputs_mat_specularity(opts, out)
         a = @spc
         @name =~ /\[\w*(s|S)\w*\]/
         fmt = "  <specularity>\n   <color r=\"%f\" g=\"%f\" b=\"%f\" />\n  </specularity>\n"
         case $1
         when 's'
            out.printf(fmt, @color.r, @color.g, @color.b)
         when 'S'
            return if a == 0
            out.printf(fmt, @color.r*a, @color.g*a, @color.b*a)
         else
            return if a == 0
            out.printf(fmt, a, a, a)
         end
      end

      def fputs_mat_glossiness(opts, out)
         a = @amb
         @name =~ /\[\w*(g|G)\w*\]/
         fmt = "  <glossiness>\n   <color r=\"%f\" g=\"%f\" b=\"%f\" />\n  </glossiness>\n"
         case $1
         when 'a'
            out.printf(fmt, @color.r, @color.g, @color.b)
         when 'A'
            return if a == 0
            out.printf(fmt, @color.r*a, @color.g*a, @color.b*a)
         else
            return if a == 0
            out.printf(fmt, a, a, a)
         end
      end

      def fputs_mat_emission(opts, out)
         a = @emi
         @name =~ /\[\w*(e|E)\w*\]/
         fmt = "  <emission>\n   <color r=\"%f\" g=\"%f\" b=\"%f\" />\n  </emission>\n"
         case $1
         when 'e'
            out.printf(fmt, @color.r, @color.g, @color.b)
         when 'E'
            return if a == 0
            out.printf(fmt, @color.r*a, @color.g*a, @color.b*a)
         else
            return if a == 0
            out.printf(fmt, a, a, a)
         end
      end
   end

   class Lights
      FMT_Location = "  <location x=\"%f\" y=\"%f\" z=\"%f\" />\n"
      FMT_Direction = "  <direction x=\"%f\" y=\"%f\" z=\"%f\" />\n"
      FMT_Color = "  <color r=\"%f\" g=\"%f\" b=\"%f\" />\n"
      attr_accessor :vertices, :visible, :root
   end
   class PointLights < Lights
      def initialize(photon, scale)
         @scale = scale
         @tag_begin = " <pointlight photon=\"#{photon}\" >"
      end
      def fputs(out, vtx, mat_id)
         return if @visible == 0

         loc = [0.0, 0.0, 0.0]
         vtx.each{|vi|
            3.times{|i|
               loc[i] += @vertices[vi][i]
            }
         }
         loc.map!{|x| x/vtx.size}

         out.puts @tag_begin
         out.printf(FMT_Location, *loc)
         col = mat_id == :default ? Color.new(@scale,@scale,@scale) : @root.materials[mat_id].color*@scale
         out.printf(FMT_Color, col.r, col.g, col.b)
         out.puts " </pointlight>"
      end
   end
   class ParallelLights < Lights
      def initialize(photon, blur, scale)
         @scale = scale
         @tag_begin = " <parallellight photon=\"#{photon}\" blur=\"#{blur}\" >"
      end
      def fputs(out, vtx, mat_id)
         return if @visible == 0
         return if vtx.size != 3

         vtx.map!{|x| Vertex.new(@vertices[x][0], @vertices[x][1], @vertices[x][2])}
         tx = [[vtx[1].distance(vtx[0]), 0],
         [vtx[2].distance(vtx[1]), 1],
         [vtx[0].distance(vtx[2]), 2]]
         tx.sort!{|a,b| b[0] <=> a[0]}
         di = vtx[tx[1][1]] - vtx[tx[0][1]]
         di = -di if ((tx[0][1] +1) % 3 == tx[1][1])

         out.puts @tag_begin
         out.printf(FMT_Direction, di.x, di.y, di.z)
         col = mat_id == :default ? Color.new(@scale,@scale,@scale) : @root.materials[mat_id].color*@scale
         out.printf(FMT_Color, col.r, col.g, col.b)
         out.puts " </parallellight>"
      end
   end
   class SphereLights < Lights
      def initialize(photon, shadowray, scale)
         @scale = scale
         @tag_begin = " <spherelight photon=\"#{photon}\" shadowray=\"#{shadowray}\" "
      end
      def fputs(out, vtx, mat_id)
         return if @visible == 0

         loc = [0.0, 0.0, 0.0]
         vtx.each{|vi|
            3.times{|i|
               loc[i] += @vertices[vi][i]
            }
         }
         loc.map!{|x| x/vtx.size}

         radius = 0.0
         vtx.each{|vi|
            y = 0.0
            3.times{|i|
               x = @vertices[vi][i] - loc[i]
               y += x * x
            }
            radius += Math.sqrt(y)
         }
         radius /= vtx.size

         out.print @tag_begin, "radius=\"#{radius}\" >\n"
         out.printf(FMT_Location, *loc)
         col = mat_id == :default ? Color.new(@scale,@scale,@scale) : @root.materials[mat_id].color*@scale
         out.printf(FMT_Color, col.r, col.g, col.b)
         out.puts " </spherelight>"
      end
   end
   class AreaLights < Lights
      FMT_VectorU = "  <vectoru x=\"%f\" y=\"%f\" z=\"%f\" />\n"
      FMT_VectorV = "  <vectorv x=\"%f\" y=\"%f\" z=\"%f\" />\n"
      def initialize(photon, shadowray, scale)
         @scale = scale
         @tag_begin = " <arealight photon=\"#{photon}\" shadowray=\"#{shadowray}\" >"
      end
      def fputs(out, vtx, mat_id)
         return if @visible == 0

         v = [0,1,2].map!{|i| @vertices[vtx[1]][i] - @vertices[vtx[0]][i] }
         u = [0,1,2].map!{|i| @vertices[vtx[2]][i] - @vertices[vtx[1]][i] }
         out.puts @tag_begin
         out.printf(FMT_Location, *@vertices[vtx[0]])
         out.printf(FMT_VectorU, *u)
         out.printf(FMT_VectorV, *v)
         col = mat_id == :default ? Color.new(@scale,@scale,@scale) : @root.materials[mat_id].color*@scale
         out.printf(FMT_Color, col.r, col.g, col.b)
         out.puts " </arealight>"
      end
   end

   class Transcoder < Loader
      # DOMo͂ꍇ̓[gIuWFNgo͂
      def document_start(ver)
         raise FormatError if ver != 1
         mqo = Document.new
         mqo.objects = ObjectChunk.new
         @out.puts '<?xml version="1.0" encoding="Shift_JIS" ?>'
         @out.puts '<redqueen>'
         return mqo
      end
      def document_end
         @out.puts '</redqueen>'
         return true
      end
      # m̃`N̏
      def chunk_start(chunk, opt)
         return nil
      end
      # DOMo͂ꍇ̓V[IuWFNgo͂
      def scene_chunk_start
         print "#{@current.class} -> scene\n" if $Debug
         return @current.scene = Scene.new
      end
      # DOMo͂ꍇ͉GIuWFNgo͂
      def backimage_chunk_start
         print "#{@current.class} -> backimage\n" if $Debug
         tmp = @current.backimages = BackImageChunk.new
         @root.scene.backimages = tmp
         return tmp
      end
      # DOMo͂ꍇ͍ގXgIuWFNgo͂
      def material_chunk_start(mat_num)
         print "#{@current.class} -> material\n" if $Debug
         return @current.materials = MaterialChunk.new
      end
      # DOMo͂ꍇ̓fIuWFNgo͂
      def object_chunk_start(obj_name)
         print "#{@current.class} -> object\n" if $Debug
         case obj_name
         when /^(?:Point|Pnt|Pt)\[\s*(\d+)\s*(?:,\s*(\d+(?:\.\d+)?)\s*)?\]/
            ret = PointLights.new($1.to_i, ($2.to_f || 1.0))
         when /^(?:Parallel|Para|Par|Pr)\[\s*(\d+)\s*,\s*(\d+(?:\.\d+)?)\s*(?:,\s*(\d+(?:\.\d+)?)\s*)?\]/
            ret = ParallelLights.new($1.to_i, $2.to_f, ($3.to_f || 1.0))
         when /^(?:Sphere|Sphe|Sph|Sp)\[\s*(\d+)\s*,\s*(\d+)\s*(?:,\s*(\d+(?:\.\d+)?)\s*)?\]/
            ret = SphereLights.new($1.to_i, $2.to_i, ($3.to_f || 1.0))
         when /^(?:Area|Ar)\[\s*(\d+)\s*,\s*(\d+)\s*(?:,\s*(\d+(?:\.\d+)?)\s*)?\]/
            ret = AreaLights.new($1.to_i, $2.to_i, ($3.to_f || 1.0))
         else
            ret = Object.new(obj_name)
            @current.objects << ret
         end
         ret.root = @root
         return ret
      end
      # DOMo͂ꍇ͔SyIuWFNgo͂
      def blob_chunk_start(opt)
         return nil
      end
      # DOMo͂ꍇ͒_XgIuWFNgo͂
      def vertex_chunk_start(vnum)
         print "#{@current.class} -> vertex\n" if $Debug
         return @current.vertices = VertexChunk.new
      end
      # DOMo͂ꍇ͒_XgIuWFNgo͂
      def vertexattr_chunk_start()
         print "#{@current.class} -> vertexattr\n" if $Debug
         return @current.vertexattr = VertexAttr.new
      end
      # DOMo͂ꍇ͒_d݃XgIuWFNgo͂
      def weit_chunk_start()
         print "#{@current.class} -> weit\n" if $Debug
         return @current.weit = Weit.new
      end
      # DOMo͂ꍇ͖ʃXgIuWFNgo͂
      def face_chunk_start(fnum)
         print "#{@current.class} -> face\n" if $Debug
         case @current
         when Lights
            return @current
         else
            tmp = FaceChunk.new
            tmp.vertices = @current.vertices
            return @current.faces = tmp
         end
      end
      # `N̏ICxg
      def chunk_end
         print "#{@parent.last.class} <- #{@current.class}\n" if $Debug
         case @current
         when Scene, Object, MaterialChunk
            @current.fputs(@opts, @out)
         end
      end
      # _f[^
      def vertex_value(x, y, z)
         @current << [x*$scale, -z*$scale, y*$scale]
      end
      # _̏d
      def weit_value(vid, weit)
         #@current << [vid, weit]
      end
      # ʃf[^
      def face_value(vnum, vtx, mat_id, uv_ary, vcol)
         return if vnum < 3 || 4 < vnum

         mat_id = :default if mat_id.nil?
         case @current
         when Lights
            @current.fputs(@out, vtx, mat_id)
         when FaceChunk
            case vnum
            when 3
               # V(a b c) --> <triangle id0="a" id1="c" id2="b" />
               v = @current.vertices
               if uv_ary.nil?
                  tmp = [
                  sprintf(FMT_Vertex, *v[vtx[0]]).intern,
                  sprintf(FMT_Vertex, *v[vtx[2]]).intern,
                  sprintf(FMT_Vertex, *v[vtx[1]]).intern]
               else
                  tmp = [
                  sprintf(FMT_VertexUV, v[vtx[0]][0], v[vtx[0]][1], v[vtx[0]][2], *uv_ary[0]).intern,
                  sprintf(FMT_VertexUV, v[vtx[2]][0], v[vtx[2]][1], v[vtx[2]][2], *uv_ary[2]).intern,
                  sprintf(FMT_VertexUV, v[vtx[1]][0], v[vtx[1]][1], v[vtx[1]][2], *uv_ary[1]).intern]
               end
               if @current.part[mat_id].nil?
                  @current.part[mat_id] = [tmp]
               else
                  @current.part[mat_id] << tmp
               end
            when 4
               v = @current.vertices
               if uv_ary.nil?
                  v0 = sprintf(FMT_Vertex, *v[vtx[0]]).intern
                  v1 = sprintf(FMT_Vertex, *v[vtx[1]]).intern
                  v2 = sprintf(FMT_Vertex, *v[vtx[2]]).intern
                  v3 = sprintf(FMT_Vertex, *v[vtx[3]]).intern
               else
                  v0 = sprintf(FMT_VertexUV, v[vtx[0]][0], v[vtx[0]][1], v[vtx[0]][2], *uv_ary[0]).intern
                  v1 = sprintf(FMT_VertexUV, v[vtx[1]][0], v[vtx[1]][1], v[vtx[1]][2], *uv_ary[1]).intern
                  v2 = sprintf(FMT_VertexUV, v[vtx[2]][0], v[vtx[2]][1], v[vtx[2]][2], *uv_ary[2]).intern
                  v3 = sprintf(FMT_VertexUV, v[vtx[3]][0], v[vtx[3]][1], v[vtx[3]][2], *uv_ary[3]).intern
               end
               if @current.part[mat_id].nil?
                  @current.part[mat_id] = [[v0, v2, v1], [v0, v3, v2]]
               else
                  @current.part[mat_id] << [v0, v2, v1]
                  @current.part[mat_id] << [v0, v3, v2]
               end
            end
         end
      end
      # ގf[^
      def material_value(mat_name, rgba, mat_hash)
         @current << Material.new(mat_name, rgba, mat_hash)
      end
      # Gf[^
      def backimage_value(view, file, opt)
         @current << BackImage.new(view, file, opt)
      end
      # xNg^̑f[^
      def attr_vector_value(atr, x, y, z)
         begin
            eval "@current.#{atr} = Vector.new(x, y, z)"
         rescue
         end
      end
      # Float^̑f[^
      def attr_float_value(atr, val)
         begin
            eval "@current.#{atr} = val"
         rescue
         end
      end
      # Int^̑f[^
      def attr_int_value(atr, val)
         begin
            eval "@current.#{atr} = val"
         rescue
         end
      end

      attr_accessor :config
      def initialize(infile, outfile, opts)
         @file = infile
         @outfile = outfile
         @opts = opts
      end

      def transcode
         open(@outfile, 'w'){|@out|
            print "\"#{@file}\" is now loading and converting... "
            ptime = Process.times.utime
            return nil if parse.nil?
            print "complete! (Elapsed time: #{Process.times.utime - ptime}s)\n"
         }
         return @root
      end
   end

   def self.transcode(infile, opts = {})
      raise FormatError unless infile =~ /\.mqo$/i
      outfile = infile.gsub(/\.mqo$/i, '.rrt')
      opts[:tracedepth] ||= 5
      opts[:gisample] ||= 16
      opts[:photonsearchradius] ||= 1
      opts[:width] ||= 640
      opts[:height] ||= 480
      opts[:oversampling] ||= 5
      opts[:bgtype] ||= 'colored'
      $scale = (opts[:scaling] ||= 0.1)
      $c_scale = (opts[:camera_scaling] ||= $scale)
      mqoLD = Transcoder.new(infile, outfile, opts)
      return mqoLD.transcode
   end
end

#= PƂŌĂяoꂽꍇ̂ݗLȏB
# CuƂČĂяoꂽꍇɂ͖B
if $0 == __FILE__
   def mqo2rrt(file_name, opts)
      if Metasequoia.transcode(file_name, opts)
         GC.start
         if opts[:do_rendering] && opts[:execute_path].is_a?(String)
            Dir.chdir(File.dirname(file_name))
            rrt_file = File.basename(file_name).gsub(/\.mqo$/i,'.rrt')
            png_file = rrt_file.gsub(/\.rrt$/,'-rq2.png')
            system(RRT[:execute_path], rrt_file, png_file, '-render')
            return true
         end
      end
      return false
   end
   def uv(u, v)
      Redqueen2::UV.new(u.to_i, v.to_i)
   end
   def xyz(x, y, z)	# Metasequoia̍Wnœ͂Rq2̍WnŕԂ
      Redqueen2::Vector.new(x.to_f, -z.to_f, y.to_f)
   end
   def rgb(r, g, b)
      Redqueen2::Color.new(r.to_f, g.to_f, b.to_f)
   end
   def parallelLight(col, dir, photon, blur)
      Redqueen2::ParallelLight.new(col, dir, photon, blur)
   end

   require 'optparse'
   # ͌ʂ邽߂Hashm
   RRT = {}
   white = rgb(1.0, 1.0, 1.0)
   gray  = rgb(0.6, 0.6, 0.6)
   black = rgb(0.0, 0.0, 0.0)

   RRT[:default_lights] = true
   begin
      load 'mqo2rq2.cfg'	# ftHgݒt@CΓǂݍ
   rescue LoadError
   end

   ARGV.options{|opts|
      opts.banner = "Usage: #{$0} [options] mqo_file ...\n"

      opts.on_head("normal options:")

      opts.on("-x", "--width=WIDTH", Integer, ":integer [640]") {|RRT[:width]|
         RRT[:width] ||= 640
      }

      opts.on("-y", "--height=HEIGHT", Integer, ":integer [480]") {|RRT[:height]|}

      opts.on("-o", "--oversampling=SAMPLE", Integer, ":integer [3]") {|RRT[:oversampling]|}

      opts.on("-d", "--tracedepth=DEPTH", Integer, ":integer [5]") {|RRT[:tracedepth]|}

      opts.on("--scaling=SCALING", Float, ":float [0.1]") {|RRT[:scaling]|
         $MQOtoRRT_Scaling = RRT[:scaling]
      }

      opts.on("--no-default-lights", "") {|RRT[:default_lights]|}

      opts.on # separator
      opts.on("background options:")

      opts.on("--bgcolor=R,G,B", Array, ":float,float,float [MQO ambient color]") {|r, g, b|
         # Array̒g`FbN
         raise OptionParser::InvalidOption unless r =~ /^(\d+\.)?\d+$/
         raise OptionParser::InvalidOption unless g =~ /^(\d+\.)?\d+$/
         raise OptionParser::InvalidOption unless b =~ /^(\d+\.)?\d+$/
         RRT[:bgcolor] = Redqueen::Vector3.new(r.to_f, g.to_f, b.to_f)
      }
      opts.on("--hdri=HDRI_FILE", String, ":*.pfm, *.png, *.ppm, *.tif") {|RRT[:hdri]|}
      opts.on("--patch=PATCH_FILE", String, ':loading patch file for mqo2rq.rb') {|pat|
         begin
            require pat
         rescue LoadError
         end
      }

      opts.on_tail # separater
      opts.on_tail("common options:")
      opts.on_tail("-?", "--help", "show this message") {puts opts; exit}
      opts.on_tail # separater
      opts.on_tail("If exists 'mqo2rq2.cfg' in current folder then preload it.")
      opts.parse!
   } or (puts "\n", ARGV.options; exit 1)

   ARGV.reject!{|x|
      !(/\.mqo$/i =~ x)
   }
   (puts ARGV.options; exit) if ARGV.empty?
   ARGV.options = nil	# no more parse

   ARGV.each{|fn|
      mqo2rrt(fn, RRT)
   }
end
